from nesbi.core.configuration import Config
from nesbi.core.helpers.logger import setup_logging
from nesbi.core.nsot import NSOTCmdb
from nesbi.core.scanner import NetworkScanner


class Nesbi(object):
    def __init__(self, config_file, **kwargs):
        self.config = Config(config_file, **kwargs)

        setup_logging(self.config.nesbi_logging_level, self.config.nesbi_logging_file,
                      self.config.nesbi_logging_to_stdout)

        self.attr_functions = kwargs.get('attr_functions', [])
        self.import_data = list()

        scanner = NetworkScanner(self.config.networks, self.config.nesbi_username,
                                 self.config.nesbi_password, self.config.nesbi_snmp_version,
                                 self.config.nesbi_snmp_community, self.config.nesbi_scan_ports,
                                 self.config.nesbi_thread_limit, self.config.nesbi_network_driver)

        self.scan_data = scanner.devices

    def generate_import_data(self):
        for location in self.scan_data:
            devices = self.scan_data.get(location)

            for device in devices:
                import_data = {}
                import_data['attributes'] = {}
                import_data['site_id'] = self.config.nsot_site_id
                import_data['hostname'] = device.get('hostname')

                try:
                    for k, v in self.config.napalm.items():
                        import_data['attributes'][k] = device.get(v)
                except AttributeError:
                    pass

                try:
                    for k, v in self.config.types.get(device.get('model')).items():
                        import_data['attributes'][k] = v
                except AttributeError:
                    pass

                try:
                    for k, v in self.config.attr_functions.items():
                        func = next(x for x in self.attr_functions if x.__name__ == v)
                        import_data['attributes'][k] = func(device)
                except AttributeError:
                    pass

                try:
                    for k, v in self.config.networks.get(location).get('attributes').items():
                        import_data['attributes'][k] = v
                except AttributeError:
                    pass

                self.import_data.append(import_data)

    def process_import_data(self):
        cmdb = NSOTCmdb(self.config.nsot_url, self.config.nsot_email, self.config.nsot_secret_key,
                        self.config.nsot_auth_header, self.config.nsot_site_id,
                        self.config.nsot_delete_objects)

        cmdb.import_data(self.import_data, self.config.nesbi_dry_run)



from nesbi.core import Nesbi


def get_monitoring_interfaces(device):
    monitoring_interfaces = list()
    interfaces_list = device.get('interface_list')
    for interface in interfaces_list:
        if device.get(interface).get('description') == "##UPLINK##":
            monitoring_interfaces.append(interface)

    return monitoring_interfaces


def main():
    config_file = 'config.yaml'

    attr_functions = [get_monitoring_interfaces]
    nesbi = Nesbi(config_file, attr_functions=attr_functions, nesbi_dry_run=True,
                  nesbi_sername="usr", nesbi_password="pw")
    nesbi.generate_import_data()
    nesbi.process_import_data()


if __name__ == "__main__":
    main()



import logging
import queue
import socket
import threading
import time
from ipaddress import IPv4Address, ip_network

import napalm

from nelsnmp.errors import SnmpError
from nelsnmp.hostinfo.device import HostInfo
from nelsnmp.snmp import SnmpHandler

from nesbi.core.helpers import deepupdate

from netmiko.exceptions import NetMikoAuthenticationException, NetMikoTimeoutException


class NetworkScanner(object):
    def __init__(self, networks, username, password, snmp_version, snmp_community,
                 scan_ports, thread_limit, network_driver):
        self.logger = logging.getLogger(__name__)
        self.networks = networks
        self.username = username
        self.password = password
        self.snmp_version = snmp_version
        self.snmp_community = snmp_community
        self.scan_ports = scan_ports
        self.thread_limit = thread_limit
        self.network_driver = network_driver
        self.devices = self._get_network_devices()

    def _get_hosts_in_network(self):
        hosts = dict()

        for k, v in self.networks.items():
            if v.get('range')[-3:] == '/32':
                hosts[k] = list([IPv4Address(v.get('range')[:-3])])

            else:
                hosts[k] = list(ip_network(v.get('range')).hosts())

                if v.get('exclude'):
                    for e in v.get('exclude'):
                        if IPv4Address(e) in hosts[k]:
                            hosts[k].remove(IPv4Address(e))

        return hosts

    def _get_reachable_devices(self):
        hosts = self._get_hosts_in_network()
        reachable_devices = dict()

        for k, v in hosts.items():
            self.logger.info(f'Start port-scan for network/location {k}')

            reachable_devices[k] = list()
            threads = list()
            q = queue.Queue()
            s = threading.Semaphore(self.thread_limit)

            for host in v:
                threads.append(threading.Thread(target=self._thread_get_reachable_devices,
                                                args=(s, str(host), q)))
                threads[-1].start()

            for t in threads:
                t.join()

            while not q.empty():
                reachable_devices[k].append(q.get())

        return reachable_devices

    def _thread_get_reachable_devices(self, s, ip, q):
        already_reachable = False

        with s:
            threading.Lock()

            for tcp_port in self.scan_ports:
                try:
                    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    sock.settimeout(0.15)
                    sock.connect((ip, tcp_port))
                    self.logger.info(f'{ip} reachable at tcp-port {tcp_port}')

                    if not already_reachable:
                        q.put(ip)
                        already_reachable = True

                except socket.error:
                    self.logger.debug(f'{ip} not reachable at tcp-port {tcp_port}')
                    pass

                sock.close()

    def _get_network_devices(self):
        hosts = self._get_reachable_devices()
        ios_devices = dict()

        for k, v in hosts.items():
            self.logger.info(f'Start device-scan for network/location {k}')
            ios_devices[k] = list()
            threads = list()
            q = queue.Queue()
            s = threading.Semaphore(self.thread_limit)

            for host in v:
                threads.append(threading.Thread(target=self._thread_get_device_facts,
                                                args=(s, host, q)))
                threads[-1].start()

            for t in threads:
                t.join()

            while not q.empty():
                ios_devices[k].append(q.get())

        return ios_devices

    def _thread_get_device_facts(self, s, host, q):
        ip = str(host)
        os = self._get_device_os(ip)

        with s:
            threading.Lock()

            if os != 'UNKNOWN':
                driver = napalm.get_network_driver(os)
                facts = dict()

                time.sleep(0.15)

                try:
                    with driver(hostname=ip, username=self.username,
                                password=self.password) as device:
                        facts = device.get_facts()
                        interfaces = device.get_interfaces()
                        interfaces_ip = device.get_interfaces_ip()
                        snmp = device.get_snmp_information()

                        first_merge = deepupdate(facts, snmp)
                        second_merge = deepupdate(interfaces, interfaces_ip)
                        final_merge = deepupdate(first_merge, second_merge)

                        q.put(final_merge)
                        self.logger.info(f"{facts.get('hostname')} scanned with napalm")

                except NetMikoAuthenticationException:
                    self.logger.error(f"{facts.get('hostname') or ip}: auth failed")

                except NetMikoTimeoutException:
                    self.logger.error(f"{facts.get('hostname') or ip}: socket-timeout")

                except napalm.base.exceptions.ConnectionClosedException as e:
                    self.logger.error(f"{facts.get('hostname') or ip}: {e}")

                except ValueError as e:
                    self.logger.error(f"{facts.get('hostname') or ip}: {e}")

    def _get_device_os(self, ip):
        dev = SnmpHandler(host=ip, version=self.snmp_version, community=self.snmp_community)
        hostinfo = HostInfo(dev)

        try:
            hostinfo.get_version()
            self.logger.info(f'{ip}: detected os "{hostinfo.os}"')
        except SnmpError:
            self.logger.error(f'{ip}: snmp-timeout')
            return False

        if self.network_driver != '':
            self.logger.info(f'{ip}: custom configs overwrites os to: "{self.network_driver}"')
            return self.network_driver

        return hostinfo.os



import json
import logging

import requests


class NSOTCmdb(object):
    def __init__(self, nsot_url, nsot_email, nsot_secret_key, nsot_auth_header,
                 nsot_site_id, nsot_delete_objects):
        self.logger = logging.getLogger(__name__)

        if nsot_secret_key:
            data = {'email': nsot_email, 'secret_key': nsot_secret_key}
            res = requests.post(f'{nsot_url}/authenticate/', data=data)
            auth_token = res.json().get('auth_token')
            headers = {'Authorization': f'AuthToken {nsot_email}:{auth_token}'}

        else:
            headers = {nsot_auth_header: nsot_email}

        headers['content-type'] = 'application/json'
        self.headers = headers
        self.nsot_url = nsot_url
        self.nsot_site_id = nsot_site_id
        self.nsot_delete_objects = nsot_delete_objects

    def import_data(self, data, dry_run):
        for d in data:
            self.logger.debug(f"data for {d.get('hostname')}:")
            self.logger.debug(d)

            if dry_run:
                self.logger.info(f"{d.get('hostname')} create dry-run")

            else:
                if self.nsot_delete_objects:
                    self._delete_data(d)

                self._create_data(d)

    def _create_data(self, d):
        response = requests.post(f'{self.nsot_url}/devices/',
                                 data=json.dumps(d), headers=self.headers, verify=False)

        status_code = response.status_code

        if status_code == 201:
            self.logger.info(f"{d.get('hostname')} created")
        elif status_code == 400:
            self.logger.warning(f"{d.get('hostname')} already in nsot or wrong attributes")
        else:
            self.logger.error(f"{d.get('hostname')} with problems, status-code:{status_code}")

    def _delete_data(self, d):
        h = d.get("hostname")
        response = requests.get(f'{self.nsot_url}/devices/{h}/',
                                headers=self.headers, verify=False)

        status_code = response.status_code
        if status_code == 200:
            response = requests.delete(f'{self.nsot_url}/sites/{self.nsot_site_id}/devices/{h}/',
                                       data=json.dumps(d), headers=self.headers, verify=False)

            status_code = response.status_code

            if status_code == 204:
                self.logger.info(f"{d.get('hostname')} deleted")
            elif status_code == 400:
                self.logger.warning(f"{d.get('hostname')} wrong attributes")
            else:
                self.logger.error(f"{d.get('hostname')} with problems, status-code: {status_code}")
        else:
            self.logger.info(f"{d.get('hostname')} not deleted (error or not existent)")



import logging.config


def setup_logging(logging_level, logging_file, logging_to_stdout):
    dict_config = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {'normal': {'format':
                                  '%(asctime)s - %(name)25s - %(levelname)8s - %(message)s'}},
        'handlers': {},
        'loggers': {},
    }

    handlers = list()

    if logging_file:
        handlers.append('info_file_handler')
        dict_config['handlers']['info_file_handler'] = {
            'class': 'logging.handlers.RotatingFileHandler',
            'level': 'NOTSET',
            'formatter': 'normal',
            'filename': logging_file,
            'maxBytes': 20971520,
            'backupCount': 10,
            'encoding': 'utf8',
        }

    if logging_to_stdout:
        handlers.append('info_console')
        dict_config['handlers']['info_console'] = {
            'class': 'logging.StreamHandler',
            'level': 'NOTSET',
            'formatter': 'normal',
            'stream': 'ext://sys.stdout',
        }

    dict_config['loggers']['nesbi'] = {'level': logging_level.upper(),
                                       'handlers': handlers}

    if handlers:
        logging.config.dictConfig(dict_config)



import os

import yaml

ATTRIBUTES = [('nesbi_username', ''),
              ('nesbi_password', ''),
              ('nesbi_dry_run', False),
              ('nesbi_network_driver', ''),
              ('nsot_url', ''),
              ('nsot_email', ''),
              ('nsot_secret_key', ''),
              ('nsot_auth_header', 'X-NSoT-Email'),
              ('nsot_site_id', '1'),
              ('nsot_delete_objects', False),
              ('nesbi_snmp_version', '2c'),
              ('nesbi_snmp_community', ''),
              ('nesbi_scan_ports', [22]),
              ('nesbi_logging_level', 'info'),
              ('nesbi_logging_file', 'nesbi.log'),
              ('nesbi_logging_to_stdout', False),
              ('nesbi_thread_limit', 5)]


class Config:
    def __init__(self, config_file, **kwargs):
        with open(config_file, 'r') as f:
            data = yaml.safe_load(f.read()) or {}

        for k, v in data.items():
            setattr(self, k, v)

        for attr in ATTRIBUTES:
            self._set_self_attribute(attr[0], attr[1], **kwargs)

    def _set_self_attribute(self, attr, default, **kwargs):
        if kwargs.get(attr):
            setattr(self, attr, kwargs.get(attr))
        elif hasattr(self, attr):
            setattr(self, attr, getattr(self, attr))
        elif attr.upper() in os.environ:
            setattr(self, attr, os.environ.get(attr.upper()))
        else:
            setattr(self, attr, default)



#!/usr/bin/env python
from setuptools import find_packages, setup


with open("requirements.txt", "r") as fs:
    reqs = [r for r in fs.read().splitlines() if (len(r) > 0 and not r.startswith("#"))]

with open("README.md", "r") as fs:
    long_description = fs.read()


__author__ = "Simon Metzger"
__author_email__ = "simon.metzger@dm.de"
__license__ = "MIT License"

__version__ = "1.0.5"

setup(name="nesbi",
      version=__version__,
      description="Nesbi (Network Scan, Build and Import data)",
      author=__author__,
      author_email=__author_email__,
      url='https://github.com/dm-drogeriemarkt/nesbi',
      include_package_data=True,
      install_requires=reqs,
      packages=find_packages(exclude=("test*", )),
      license=__license__,
      long_description=long_description,
      long_description_content_type="text/markdown",
      test_suite="tests",
      platforms="any",
      classifiers=["Development Status :: 4 - Beta",
                   "Programming Language :: Python :: 3",
                   "Programming Language :: Python :: 3 :: Only",
                   "Programming Language :: Python :: 3.7",
                   "Programming Language :: Python :: 3.8",
                   "Programming Language :: Python :: 3.9",
                   "Programming Language :: Python :: 3.10",
                   "License :: OSI Approved :: MIT License"
                   ])
